/**
 * Copyright 2005 Simon Lehmayr
 *
 * This file is part of Asteroid Zone. Asteroid Zone is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 * Asteroid Zone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Asteroid Zone; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 */

package asteroids;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.util.*;
import javax.microedition.lcdui.game.*;

/**
 * The game main screen. The class also implements the game's main loop.
 * @author Jean-Francois Doue
 * @author Pavel Machek
 * @author Simon Lehmayr
 * @version 1.6, 2005/12/20
 */
public class Field extends GameCanvas implements Runnable {
    /** Thread used for key handling and animation */
    private Thread thread;

    /** Game difficulty */
    public static final int EASY = 0;
    public static final int NORMAL = 1;
    public static final int HARD = 2;
    public static final String DIFFSTRING[] = {"easy", "normal", "hard"};
    public static int difficulty = 1;
    public static final int BONUSLIFE[] = {500, 1000, 1500};
    public static final int GAME_SPEED[] = {60, 50, 40};
    public static final int MAX_LEVEL = 10;

    public static final int UFOSCORE[] = {1000, 250, 215};

    /** Startup recognition */
    public static boolean firstTime = true;

    /** Level delaying */
    public static final int LEVELDELAY = 30;
    public static int _nextLevelDelay = 0;

    /** The 'title' game state. Displays the game name and author. */
    public static final byte TITLE_STATE = 0;
    
    /** The 'control' game state. Displays the mapping between keys and game commands. */
    public static final byte CONTROL_STATE = 1;
    
    /** The 'points' game state. Displays the value of each screen items. */
    public static final byte POINTS_STATE = 2;
    
    /** The 'high score' game state. Displays the high score table. */
    public static final byte HISCORE_STATE = 3;
    
    /** The 'game over' game state. Displays the 'game over' message. */
    public static final byte GAMEOVER_STATE = 4;
    
    /** The 'new high score' game state. Lets players enter their name. */
    public static final byte NEWHIGHSCORE_STATE = 5;
    
    /** The 'game' game state. The player is actively playing the game. */
    public static final byte GAME_STATE = 6;

    /** When not in game mode, the game switches its state every TIMEOUT ms */
    public static final long TIMEOUT = 8000L;

    /** Font setup */
    public static Font smallFont = Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_PLAIN, Font.SIZE_SMALL);
    public static Font bigFont = Font.getFont(Font.FACE_PROPORTIONAL, Font.STYLE_BOLD, Font.SIZE_LARGE);

/* Old color scheme
    public static final int COLOR_BACKGROUND = 0x00FFFFFF;
    public static final int COLOR_FRAME = 0x00000000;
    public static final int COLOR_ASTEROID = 0x000000FF;
    public static final int COLOR_EXPLOSION = 0x00FF00FF;
    public static final int COLOR_ROCKET = 0x00FF0000;
    public static final int COLOR_FONT = 0x00000000;
*/
    /** Color scheme */
    public static final int COLOR_BACKGROUND = 0x00000000;
    public static final int COLOR_FRAME = 0x00000000;
    public static final int COLOR_ASTEROID = 0x00FFFFFF;
    public static final int COLOR_EXPLOSION = 0x00FFA080;
    public static final int COLOR_ROCKET = 0x00FFFFFF;
    public static final int COLOR_UFO = 0x00A0FFB0;
    public static final int COLOR_MISSILE = 0x00FFFF80;
    public static final int COLOR_SHIP = 0x00FFFFFF;
    public static final int COLOR_SHIP_INV = 0x00606060;
    public static final int COLOR_SHIELD = 0x0050B080;
    public static final int COLOR_FONT = 0x00D0D0D0;

    /** Color blink for invincible ship */
    private boolean colorToggle=false;

    /** Time spent moving objects and computing collisions. */
    //public static long computeavg;
    
    /** Time spent painting screens. */
    //public static long paintavg;
    
    /** Number of game frames since beginning. */
    //public static long frames;

    private byte _state;                        // The current game state
    private Slideshow _nextSlide;                // A TimeTask used to trigger the transition to the next state.
    private int _lives;                                // Number of lives left.
    private int _nextBonusLife;                // Score above which next extra life will be awarded.
    private int _score;                                // Current score.
    public byte _level;                        // Current level.
    private boolean _paused;                // Set the true by the Game if it wants to enter its paused state
                                                            // This causes the main game thread to exit.
    private boolean _frozen;                        // Set when another Displayable is being displayed by the game
                                                            // (nothing happens in the game in between, but the game 
                                                            // must be able to resume).
    private boolean _ufoActive;
    private int _nextUfoScore;

    /** Various texts for display */
    private char[] _scoreString;                // Score string
    private char _liveString;                        // Lives string
    private char[] _diffString;                // Difficulty string
    private byte _diffStringLength;
    private char[] _levelString                // The current level indicator
     = {'L', 'e', 'v', 'e', 'l', ' ', '0'};
    private byte _levelStringTimer;         // The number of frames during which the level indicator is displayed
    private int _liveStringWidth;                // In pixels.
    private int _lastKeyPressed;                // The last key pressed by the player.
    private boolean _isRepeatedKey;        // True if _lastKeyPressed was obtained through keyRepeated().

    /** Game field constructor */
    public Field() {
        super(false);
//        System.out.println("Field :: <construct>");
        //computeavg = 0;
        //paintavg = 0;
        //frames = 0;

        // Determine the dimensions of the Canvas.
        // The game has been developed using the j2mewtk's
        // DefaultColorPhone, using a resolution of 100x96.
        // Simon developes on a Sony Ericsson K750i
        // Object are sized proportionaly on other platforms
        // ratioDenom.
        Mobile.width = getWidth();
        Mobile.height = getHeight();
        Mobile.ratioNum = Math.min(Mobile.width, Mobile.height);
       
        // The score is stored as a char array for greater
        // efficiency.
        _scoreString = new char[4];
        _ufoActive = false;

        // Set game difficulty to normal
        setDifficulty(1);
    }

    /**
     * The canvas is being displayed.
     * Stop the event handling and animation thread.
     */
    protected void showNotify() {
//        System.out.println("Field :: showNotify");
        if (firstTime) {
            thread = new Thread(this);
            thread.start();
            firstTime=false;
        }
    }

    /**
     * The canvas is being removed from the screen.
     * Stop the event handling and animation thread.
     */
    protected void hideNotify() {
//        System.out.println("Field :: hideNotify");
//        thread = null;
    }

    /**
     * Starts a new game.
     */
    public final void newGame() {
        _level = 0;
        setLives(4);
        _nextBonusLife = BONUSLIFE[difficulty];
        setScore(0);
        Asteroid.asteroids.removeAll();
        _lastKeyPressed = 0;
        _isRepeatedKey = false;
        Ship.ship.reset();
        Ufo.ufo.reset();
        Ship.ship.shields = Ship.ship.SHIELD_TIME[difficulty];
        Missile.missile.reset();
        _nextUfoScore = UFOSCORE[difficulty];
    }

    /**
     * Returns the number of remaining ships.
     */
    public final int getLives() {
        return _lives;
    }

    /**
    * Sets the number of remaining ships.
    */
    public final void setLives(int lives) {
        if (lives < 0) {
            lives = 0;
        }
        _lives = lives;
        _liveString = (char)('0' + _lives);
        _liveStringWidth = smallFont.charWidth(_liveString);
    }

    /**
     * Sets the game difficulty.
     */
    public final void setDifficulty(int diff) {
        if (diff < 0) {
            diff = 0;
        }
        difficulty = diff;
        _diffString = DIFFSTRING[diff].toCharArray();
        _diffStringLength = (byte)DIFFSTRING[diff].length();
    }

    /**
     * Returns the current score.
     */
    public final int getScore() {
        return _score;
    }

    /**
     * Updates the score.
     */
    public final void setScore(int score) {
        _score = score;

        // Convert the score to an array of chars
        // of the form: 0000
        Scores.toCharArray(_score, _scoreString);

        // Extra lives are awarded every nnn points.
        if (_score >= _nextBonusLife) {
            _nextBonusLife += BONUSLIFE[difficulty];
            setLives(_lives + 1);
        }
        if (_score >= _nextUfoScore) {
            _nextUfoScore += UFOSCORE[difficulty];
            Ufo.ufo.startUfo();
        }
    }

    /**
     * Initializes a new level.
     */
    private final void _nextLevel() {
        // At most 7 large asteroids per screen.
        _nextLevelDelay = LEVELDELAY;
        if (_level < MAX_LEVEL) {
            _level++;
        }
        _ufoActive = false;
	    if (difficulty != EASY && _level > 1) {
            _ufoActive = true;
            Ufo.ufo.reset();
            Missile.missile.reset();
        }

        // Populate the game level with large asteroids.
        for (byte i = 0; i < 3 + _level; i++) {
            Asteroid asteroid = (Asteroid)Asteroid.asteroids.addNewObject();
            if (asteroid != null) {
                Asteroid.randomInit(asteroid);
            }
        }

        // Clear all the existing rockets or explosions.
        Rocket.rockets.removeAll();
        Explosion.explosions.removeAll();

        if (_level >= 1) {
            _levelString[6] = (char)('0' + _level);
            _levelStringTimer = 25;
        }
    }

    /**
     * Overriden from Canvas.
     */
    public void paint(Graphics g) {
        flushGraphics();
    }

    /**
     * Overriden from Canvas.
     */
    protected void keyPressed(int keyCode) {
            _lastKeyPressed = keyCode;
    }

    protected void keyRepeated(int keyCode) {
            _lastKeyPressed = keyCode;
            _isRepeatedKey = true;
    }
    
    /**
     * Returns the current state.
     */
    public final byte getState() {
        return _state;
    }

    /**
     * Triggers a transition from one state of the game automaton
     * to the other.
     */
    public final void setState(byte newState) {
        switch (newState) {
            case TITLE_STATE:
                _nextSlide = new Slideshow(CONTROL_STATE);
                Game.timer.schedule(_nextSlide, TIMEOUT);
                break;
            case CONTROL_STATE:
                _nextSlide = new Slideshow(POINTS_STATE);
                Game.timer.schedule(_nextSlide, TIMEOUT);
                break;
            case POINTS_STATE:
                _nextSlide = new Slideshow(HISCORE_STATE);
                Game.timer.schedule(_nextSlide, TIMEOUT);
                break;
            case HISCORE_STATE:
                _nextSlide = new Slideshow(TITLE_STATE);
                Game.timer.schedule(_nextSlide, TIMEOUT);
                break;
            case GAMEOVER_STATE:
                if (Game.scores.isHighScore(_score)) {
                    _nextSlide = new Slideshow(NEWHIGHSCORE_STATE);
                    Game.timer.schedule(_nextSlide, TIMEOUT / 2);
                } else {
                    _nextSlide = new Slideshow(HISCORE_STATE);
                    Game.timer.schedule(_nextSlide, TIMEOUT);
                }
                break;
            case NEWHIGHSCORE_STATE:
                    setFrozen(true);
                if (Game.display.getCurrent() != Game.scoreForm) {
                    Game.enterHighScore();
                }
                break;
            case GAME_STATE:
                if (_nextSlide != null) {
                    _nextSlide.cancel();
                    _nextSlide = null;
                }
                break;
        }
        _state = newState;
    }

    /**
     * Asks the game's main thread to exit.
     */
    public final void pause() {
//        System.out.println("Field :: pause");
        _paused = true;
    }
    
    /**
     * Freezes the game
     */
    public final void setFrozen(boolean frozen) {
        _frozen = frozen;
        if (_frozen) {
            if (_nextSlide != null) {
                _nextSlide.cancel();
                _nextSlide = null;
            }
        }
    }

    /**
     * Executes the game's main loop.
     */
    public void run() {
      Thread mythread = Thread.currentThread();
      while (mythread == thread) {
        try {
            while (!_paused) {
                long time1 = System.currentTimeMillis();
                //long time2;
                long time3;
                long elapsed = 0;
                if (!_frozen) {

                                        // If the field has been cleared, change the level.
                    if (Asteroid.asteroids.size() == 0) {
                        if (_nextLevelDelay==0) _nextLevel(); else _nextLevelDelay--;
                    }

                    // Animate the explosion and remove those which are done.
                    Explosion.explode();
                                   
                    // Move the asteroids.
                    Asteroid.move();

                    // Move the rockets and remove those which have expired.
                    Rocket.move();
		    
		            // Move the ufo (if active)
		            if (_ufoActive) {
                        Ufo.ufo.move();
                        Missile.missile.move();
                    }

                    // Handle user events
                    // Move the ship (only while the game is actually proceeding)
                    if (_state == GAME_STATE) {
                        Ship.ship.isShieldActive = false;
                        int keyState = getKeyStates();
                        if ((keyState & LEFT_PRESSED) != 0) Ship.ship.rotate(-1);
                        else if ((keyState & RIGHT_PRESSED) != 0) Ship.ship.rotate(1);
                        else if ((keyState & FIRE_PRESSED) != 0) Ship.ship.shoot(Rocket.rockets);
                        else if ((keyState & UP_PRESSED) != 0) Ship.ship.burn();
                        else if ((keyState & DOWN_PRESSED) != 0) Ship.ship.brake();
                        else if ((keyState & GAME_A_PRESSED) != 0) Ship.ship.shieldUp();
                        else if ((keyState & GAME_B_PRESSED) != 0) Ship.ship.teleport();
                        
                        Ship.ship.move();
                    }

                    // Compute collisions between the asteroids and
                    // the ship and the rockets.
                    Asteroid.collisionDetection();
                    if (_ufoActive) {
                        Ufo.collisionDetection();
                        Missile.collisionDetection();
                    }

					// Detect game over
                    if ((_state == GAME_STATE) && (_lives <= 0) && (Ship.ship.isAlive)) {
                        setState(GAMEOVER_STATE);
                    }

                    Graphics gr = getGraphics();
                    gr.setColor(COLOR_BACKGROUND);
                    gr.fillRect(0, 0, getWidth() - 1, getHeight() - 1);
                    gr.setColor(COLOR_FRAME);
                    gr.drawRect(0, 0, getWidth() - 1, getHeight() - 1);
        
                     // Draw the asteroids
                    gr.setColor(COLOR_ASTEROID);
                    Asteroid.draw(gr);

                    // Draw the explosions
                    gr.setColor(COLOR_EXPLOSION);
                    Explosion.draw(gr);

                    // Draw the rockets
                    gr.setColor(COLOR_ROCKET);
                    Rocket.draw(gr);
        
                     // Draw the ufo & missile
                    if (_ufoActive) {
                        gr.setColor(COLOR_UFO);
                        Ufo.ufo.draw(gr);
                        gr.setColor(COLOR_MISSILE);
                        Missile.missile.draw(gr);
                    }

                    switch(_state) {
                        case TITLE_STATE:
                            Slideshow.drawTitleScreen(gr);
                            break;
                
                        case CONTROL_STATE:
                            Slideshow.drawControlScreen(gr);
                            break;
                
                        case POINTS_STATE:
                            Slideshow.drawPointsScreen(gr);
                            break;

                        case HISCORE_STATE:
                            Slideshow.drawHighScoresScreen(gr);
                            break;
                
                        case GAME_STATE:
                            gr.setColor(COLOR_FONT);
                            //gr.drawRect(10, 10, 16, 16);

                            gr.setFont(smallFont);
                            gr.drawChars(_scoreString, 0, 4, 2, getHeight() - 1, Graphics.BOTTOM|Graphics.LEFT);
                            gr.drawChars(_diffString, 0, _diffStringLength, getWidth()/2, getHeight() - 1, Graphics.BOTTOM|Graphics.HCENTER);
                            gr.drawChar(_liveString, Mobile.width - _liveStringWidth - 1,  Mobile.height, Graphics.BOTTOM|Graphics.LEFT);
            
                            // Display the level indicator (only during the first few frames of the level)
                            if (_levelStringTimer > 0) {
                                _levelStringTimer--;
                                int x = (Mobile.width - Field.smallFont.charsWidth(_levelString, 0, _levelString.length)) >> 1;
                                int y = ((Mobile.height) >> 1) + Field.smallFont.getHeight();
                                gr.drawChars(_levelString, 0, _levelString.length, x, y, Graphics.BOTTOM|Graphics.LEFT);
                            }
                            gr.setColor(COLOR_SHIELD);
                            gr.drawLine(30, getHeight() - 3, 30, getHeight() - 12);
                            gr.drawLine(31, getHeight() - 5, 59, getHeight() - 5);
                            gr.fillRect(30, getHeight() - 9, Ship.ship.shields * 30 / Ship.ship.SHIELD_TIME[difficulty], 5);
                            gr.drawLine(60, getHeight() - 3, 60, getHeight() - 12);
                            if (Ship.ship.isInvincible) { if (colorToggle) gr.setColor(COLOR_SHIP); else gr.setColor(COLOR_SHIP_INV); colorToggle=!colorToggle;}
                            else gr.setColor(COLOR_SHIP);
                            Ship.ship.draw(gr);
                            gr.setColor(COLOR_SHIP);
                            boolean tempShield = Ship.ship.isShieldActive;
                            Ship.ship.isShieldActive = false;
                            Ship.draw(0, Mobile.width - Ship.radius - 2 - _liveStringWidth, Mobile.height - Ship.radius - 1, gr);
                            Ship.ship.isShieldActive = tempShield;
                            break;
                
                        case GAMEOVER_STATE:
                            Slideshow.drawGameOverScreen(gr);
                            break;
                    }
                    if (mythread == thread) flushGraphics();

                    // Determine the time spent to draw the frame.
                    time3 = System.currentTimeMillis();
                    //paintavg += (time3 - time2);
                    //frames++;

                    // Determine the total time for the frame.
                    elapsed = time3 - time1;
                }
                // Sleep for a while (at least 20ms)
                try {
                    Thread.currentThread().sleep(Math.max(GAME_SPEED[difficulty] - elapsed, 20));
                } catch(java.lang.InterruptedException e) {}
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
      }
    }
}
