/**
 * Copyright 2005 Simon Lehmayr
 *
 * This file is part of Asteroid Zone. Asteroid Zone is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 * Asteroid Zone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Asteroid Zone; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 */

package asteroids;

import javax.microedition.lcdui.*;
import java.util.*;

/**
 * Class to implement the homing missile.
 * It can rotate and moves towards the player's ship.
 * @author Simon Lehmayr
 * @version 1.6, 2005/12/20
 */
public class Missile extends Mobile {
    public boolean isAlive, isActive;

    /** The orientation of the missile. */
    public byte angle;

    /** Difficulty dependent values */
    private static final int VALUEMISSILE[] = {20, 40, 60};
    private static final int SPEEDMISSILE[] = {256*256, 768*768, 1024*1024};
    private static final int FUELMISSILE[] = {100, 500, 1500};
    
    /** Constants */
    private static final int THRUSTHOLDOFF = 4;
    private static final int HOMINGHOLDOFF = 2;
    private static final int SOUNDHOLDOFF = 30;

    /** The coordinates of an array of points representing the missile at various orientations. */
    public static byte[] xcoords;
    public static byte[] ycoords;
    
    /** The size of the missile dependent on the angle. */
    public static byte radii[] = {//new byte[32];
        7, 5, 5, 6, 4, 2, 2, 2,
        2, 2, 2, 2, 2, 2, 2, 2,
        2, 4, 5, 3, 2, 2, 2, 2,
        2, 2, 2, 2, 2, 2, 2, 3};

    /** Various counters */
    private int holdOffThrust, holdOffHoming, holdOffMissileSound, fuel;
    
    /** The missile singleton */
    public static Missile missile;

    private byte _explosionFrame;

    static {
        // Instantiate the missile
        missile = new Missile();
        
        final byte[] pointx = {-3,-2, -3, 3};
        final byte[] pointy = {-1, 0, 1, 0};

/*        System.out.println("1");
        // Calculate all collision detection radii (based on given missile shape)
        int i, a=0, b=0, c=0, alpha=0, beta=0, gamma=0, alpha_max=0, n=0, x1=0, y1=0, x2=0, y2=0;
        for (i = 0; i < 32; i++) {
            if (i >= alpha_max) {
                x1 = pointx[n]; y1 = pointy[n++]; if (n<4) {x2 = pointx[n]; y2 = pointy[n];} else {x2 = pointx[0]; y2 = pointy[0];}
                a = distance(x1, y1, x2, y2) ; b = distance(0, 0, x2, y2); c = distance(0, 0, x1, y1);
                alpha_max += acos(256*(b*b+c*c-a*a)/(2*b*c));
                beta = acos(256*(a*a+c*c-b*b)/(2*a*c));
            }
            alpha = i - alpha_max;
            gamma = 16 - alpha - beta; if (gamma < 0) gamma += 32;
            radii[i] = (byte)(c * sin[beta] * ratioNum / sin[gamma] / ratioDenom);
        }
        System.out.println("2");*/
        
        // Calculate all missile point coordinates for all possible angles
        xcoords = new byte[128];
        ycoords = new byte[128];
        for (int i = 0; i < 32; i++) {
            int offset = i << 2;
            for (int j = 0; j < 4; j++) {
                xcoords[offset + j] = (byte)(((pointx[j] * cos[i] - pointy[j] * sin[i]) * ratioNum / ratioDenom) >> 6);
                ycoords[offset + j] = (byte)(((pointx[j] * sin[i] + pointy[j] * cos[i]) * ratioNum / ratioDenom) >> 6);
            }
        }
    }

    /** Empty constructor */
    public Missile() {
    }

    /**
     * Positions the missile at its original location, orientation
     * and velocity.
     */
    public final void reset() {
        Field field = Game.field;
	    Ufo ufo = Ufo.ufo;
        Ship ship = Ship.ship;
        if (ufo.Direction == -1)
        {
            angle = 0;
            moveTo(ufo.x + 10, ufo.y);
        } else
        {
            angle = 16;
            moveTo(ufo.x - 10, ufo.y);
        }
        vx = 0;
        vy = 0;
        holdOffMissileSound = 0;
        holdOffThrust = THRUSTHOLDOFF;
        holdOffHoming = HOMINGHOLDOFF;
        fuel = FUELMISSILE[field.difficulty];
        isAlive = true;
        isActive = false;
        rotate(angle);
    }

    /**
     * Called when the missile is shot down.
     */
    public final void explode() {
        isAlive = false;
        _explosionFrame = 0;
    }

    /** Called when missile is fired from the UFO */
    public final void fire() {
        if (isActive) return;
        reset();
        isActive = true;
    }

    /**
     * Move the missile to its next position.
     */
    public final void move() {
      Field field = Game.field;
      Soundsystem sound = Game.sound;
      if (isActive) 
      {
        if (isAlive) {
            Ship ship = Ship.ship;
            if (fuel > 0) fuel--; else {/*explode();*/ return;}
            if (holdOffMissileSound == 0) {sound.playSound(sound.MISSILE); holdOffMissileSound = SOUNDHOLDOFF;} else holdOffMissileSound--;
            if (holdOffThrust == 0) {
                burn();
                holdOffThrust = THRUSTHOLDOFF;
            } else holdOffThrust--;
            if (holdOffHoming == 0) {
/*                byte dest_angle = (byte)(getAngle(ship.x, ship.y, x, y)*2);
                byte d_angle = (byte)(((angle - (dest_angle+16) % 32))); //System.out.println("d_angle="+(d_angle*180/32));
//                if (d_angle < 10) {rotate(2); brake();}
//                else if (d_angle > 10) {rotate(-2); brake();}
                if (d_angle < 0) rotate(1);
                else if (d_angle > 0) rotate(-1);
                holdOffHoming = HOMINGHOLDOFF;
*/
                byte dest_angle=0;
                int dx = x - ship.x;
                int dy = y - ship.y;
                int dist = (int)Math.sqrt(dx*dx + dy*dy);
                if (dx >= 0 && dy >= 0) {dest_angle = (byte)(asin[dy*256/dist]/3938); /*System.out.println("dx="+dx+", dy="+dy+", dist="+dist+", alpha="+(asin[dy*256/dist]/700));*/ }
                else if (dx >= 0 && dy < 0) {dest_angle = (byte)(32 - asin[-dy*256/dist]/3938); /*System.out.println("dx="+dx+", dy="+dy+", dist="+dist+", alpha = "+(360-asin[-dy*256/dist]/700));*/ }
                else if (dx < 0 && dy >= 0) {dest_angle = (byte)(16 - asin[dy*256/dist]/3938); /*System.out.println("dx="+dx+", dy="+dy+", dist="+dist+", alpha = "+(180-asin[dy*256/dist]/700));*/ }
                else if (dx < 0 && dy < 0) {dest_angle = (byte)(16 + asin[-dy*256/dist]/3938); /*System.out.println("dx="+dx+", dy="+dy+", dist="+dist+", alpha = "+(180+asin[-dy*256/dist]/700));*/ }
                byte d_angle = (byte)(((angle - (dest_angle+16) % 32))); //System.out.println("d_angle="+(d_angle*180/32));
                if (d_angle < 0) rotate(1);
                else if (d_angle > 0) rotate(-1);
                holdOffHoming = HOMINGHOLDOFF;
            } else holdOffHoming--;
            xold = x;
            yold = y;
            _x += vx;
            _y += vy;
            x = _x >> 8;
            y = _y >> 8;
            
            // If a border has been hit, wrap the trajectory around
            // the screen. The new origin is the projection of the
            // intersection point on the opposite border.
            if (x <= 0) {
                moveTo(width - 2, y);
            } else if (x >= width - 1) {
                moveTo(1, y);
            } else if (y <= 0) {
                moveTo(x, height - 2);
            } else if (y >= height - 1) {
                moveTo(x, 1);
            }
         } else {
            _explosionFrame++;
            if (_explosionFrame > 20) {
                reset();
            }
        }
      }
    }

    /**
     * Changes the orientation of the missile.
     */
    public final void rotate(int direction) {
        if (isAlive) {
            angle += direction;
            if (angle > 31) {
                angle -= 32;
            } else if (angle < 0) {
                angle += 32;
            }
        }
    }

    /**
     * Adds velocity to the missile.
     */
    public final void burn() {
        if (isAlive) {
            Field field = Game.field;
            int newvx = vx + (cos[angle] << 1);
            int newvy = vy + (sin[angle] << 1);
            // clamp the velocity.
            if (newvx * newvx + newvy * newvy < SPEEDMISSILE[field.difficulty]) {
                vx = newvx;
                vy = newvy;
            }
        }
    }

    /**
     * Decreases velocity of the missile.
     */
    public final void brake() {
        if (isAlive) {
            Field field = Game.field;
            int newvx = vx - (cos[angle] << 1);
            int newvy = vy - (sin[angle] << 1);
            // clamp the velocity.
            if (newvx * newvx + newvy * newvy < SPEEDMISSILE[field.difficulty]) {
                vx = newvx;
                vy = newvy;
            }
        }
    }

    /**
     * Compute collisions between the missile and
     * the ship and the rockets.
     */
    public static final void collisionDetection() {
        Field field = Game.field;
        Ship ship = Ship.ship;
        Soundsystem sound = Game.sound;
        if (!missile.isActive || !missile.isAlive) return;
            
        // Collision detection makes sense only while the game is
        // being played.
        if ((field != null) && (field.getState() == Field.GAME_STATE)) {
            
            int score = field.getScore();
            int newScore = score;
            Pool rockets = Rocket.rockets;
            Pool explosions = Explosion.explosions;
            
            // Detect ship - missile collisions.
            if (!ship.isInvincible && ship.isAlive) {
                if (!ship.isShieldActive) {
                    int offset = ship.angle << 2;
                    boolean hit = false;
                    for (int i = 0; i < 4; i++) {   // Check all ship points for collision with missile
                        int shipx = ship.x + Ship.xcoords[offset + i], shipy = ship.y + Ship.ycoords[offset + i];
                        int d = distance(shipx, shipy, missile.x, missile.y);
                        int alpha = getAngle(missile.x, missile.y, shipx, shipy, d);
                        if (d <= Missile.radii[(alpha-missile.angle+32)%32]) {
                            int lives = field.getLives();
                            if (lives > 0 && !hit) {
                                sound.playSound(sound.CRASH);
                                ship.explode();
                                missile.explode();
                                field.setLives(lives - 1);
                                hit = true;
                                break;
                            }
                        }
                    }
                    offset = missile.angle << 2;
                    if (!hit) for (int i = 0; i < 4; i++) {   // Check all missile points for collision with ship
                        int missilex = missile.x + Missile.xcoords[offset + i], missiley = missile.y + Missile.ycoords[offset + i];
                        int d = distance(missilex, missiley, ship.x, ship.y);
                        int alpha = getAngle(ship.x, ship.y, missilex, missiley, d);
                        if (d <= Ship.radii[(alpha-ship.angle+32)%32]) {
                            int lives = field.getLives();
                            if (lives > 0 && !hit) {
                                sound.playSound(sound.CRASH);
                                ship.explode();
                                missile.explode();
                                field.setLives(lives - 1);
                                hit = true;
                                break;
                            }
                        }
                    }
                }
                else {
                    int offset = missile.angle << 2;
                    for (int i = 0; i < 4; i++) {   // Check all missile points for collision with ship shields
                        int missilex = missile.x + Missile.xcoords[offset + i], missiley = missile.y + Missile.ycoords[offset + i];
                        if (distance(missilex, missiley, ship.x, ship.y) <= ship.SHIELD_RADIUS) {
                            sound.playSound(sound.EXPLOSION);
                            missile.explode();
                            ship.shields -= VALUEMISSILE[field.difficulty]; if (ship.shields < 0) ship.shields = 0;
                            break;
                        }
                    }
                }
            }

            // Detect missile - rocket collisions.
            if (missile.isAlive) for (rockets.current = rockets.count - 1; rockets.current >= 0;) {
                Rocket r = (Rocket)rockets.pool[rockets.current--];
                for (int i = 0; i < 4; i++) {
                    int d = distance(missile.x, missile.y, r.x, r.y);
                    int alpha = getAngle(missile.x, missile.y, r.x, r.y, d);
                    if (d <= Missile.radii[(alpha-missile.angle+32)%32]) {
                        rockets.removeCurrent();
                        newScore += VALUEMISSILE[field.difficulty];
                        missile.explode();
                        sound.playSound(sound.EXPLOSION);
                        break;
                    }
                }
            }

            if (newScore != score) {
                field.setScore(newScore);
            }
        }
    }
         
    /**
     * Draws the missile at the specified location and orientation.
     */
    public static void draw(int orientation, int xpos, int ypos, Graphics g) {
        int offset = orientation << 2;
        g.drawLine(xcoords[offset] + xpos, ycoords[offset] + ypos,
                   xcoords[offset + 1] + xpos, ycoords[offset + 1] + ypos);
        g.drawLine(xcoords[offset + 1] + xpos, ycoords[offset + 1] + ypos,
                   xcoords[offset + 2] + xpos, ycoords[offset + 2] + ypos);
        g.drawLine(xcoords[offset + 2] + xpos, ycoords[offset + 2] + ypos,
                   xcoords[offset + 3] + xpos, ycoords[offset + 3] + ypos);
        g.drawLine(xcoords[offset + 3] + xpos, ycoords[offset + 3] + ypos,
                   xcoords[offset + 0] + xpos, ycoords[offset + 0] + ypos);
    }

    /**
     * Draws the missile in the specified graphic context.
     */
    public final void draw(Graphics g) {
        if (!isActive) return;
        if (isAlive) {
            draw(angle, x, y, g);
        } else {
            if (_explosionFrame < 6) {
                g.setColor(Field.COLOR_EXPLOSION);
                int radius = _explosionFrame;
                g.drawArc(x - radius, y - radius, radius << 1, radius << 1, 0, 360);
                radius = _explosionFrame + 2;
                g.drawArc(x - radius, y - radius, radius << 1, radius << 1, 0, 360);
                radius = _explosionFrame + 4;
                g.drawArc(x - radius, y - radius, radius << 1, radius << 1, 0, 360);
            }
        }
    }
}

