/**
 * Copyright 2005 Simon Lehmayr
 *
 * This file is part of Asteroid Zone. Asteroid Zone is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 * Asteroid Zone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Asteroid Zone; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 */

package asteroids;

/**
 * Class to implement efficient, pool based collections.
 * class supports an iterator method for collection traversal,
 * and methods to add and remove elements. collection size
 * cannot grow beyond size of initial pool of objects
 * provided for collection.
 * @author Jean-Francois Doue
 * @author Simon Lehmayr
 * @version 1.5, 2005/12/17
 */
public class Pool extends Object {
    /** An object array pool where objects are actually stored. */
    public Object[] pool;

    /** number of objects currently in pool. */
    public int count;

    /** index of current object. */
    public int current;

    /**
     * Initializes collection with a pool of instances.
     */
    public Pool(Object[] pool) {
        this.pool = pool;
    }

    /**
     * Resets collection iterator.
     */
    public final void reset() {
        current = count - 1;
    }

    /**
     * Returns next object in collection.
     */
    public final Object next() {
        if (current >= 0) {
            return pool[current--];
        }
        return null;
    }

    /**
     * Removes current object from collection.
     */
    public final void removeCurrent() {
        if (current + 1 < count - 1) {
            Object tmp = pool[current + 1];
            pool[current + 1] = pool[count - 1];
            pool[count - 1] = tmp;
        }
        count--;
    }

    /**
     * Adds a new object to collection.     
     * returned object must be initialized by callee.
     * Returns null if collection capacity has been
     * exceeded.
     */
    public final Object addNewObject() {
        if (count >= pool.length) {
            return null;
        }
        return pool[count++];
    }

    /**
     * Removes all objects from collection.
     */
    public final void removeAll() {
        count = 0;
    }

    /**
     * Returns size of collection.
     */
    public final int size() {
        return count;
    }
}
