/**
 * Copyright 2005 Simon Lehmayr
 *
 * This file is part of Asteroid Zone. Asteroid Zone is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 * Asteroid Zone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Asteroid Zone; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 */

package asteroids;

import javax.microedition.lcdui.*;
import java.util.*;

/**
 * Class to implement the player's spaceship. The spaceship
 * can rotate, move, shoot and teleport itself.
 * @author Jean-Francois Doue
 * @author Simon Lehmayr
 * @version 1.6, 2005/12/20
 */
public class Ship extends Mobile {

    public boolean isAlive, isInvincible, isShieldActive;   // Ship states
    public int countInvincible;               // Counter going down when invincible

    /** Difficulty level dependent values */
    private static final int INVINCIBLE_TIME[] = {100, 70, 40};     // The time (in frames) a fresh ship is invincible
    private static final byte FIRESPEED[] = {3, 3, 6};                 // The time between bullets ("rockets")
    private static final byte BURSTLENGTH[] = {8, 5, 3};            // The number bullets shot in a row ("burst")
    private static final byte BURSTDELAY[] = {3, 10, 15};          // The delay between shot bursts
    public static final int SHIELD_TIME[] = {500, 150, 200};    // The time the shield can be used in total

    /** Teleport restriction counter */
    private int teleportCounter;
    private static final int TELEPORT_TIME[] = {200, 500, 1000};

    /** The orientation of the ship. */
    public byte angle;

    /** The coordinates of an array of points representing the ship at various orientations. */
    public static byte[] xcoords;
    public static byte[] ycoords;
    public static int radius;

    /** The size of the ship (in radii). */
    public static byte radii[] = {
        9, 5, 4, 3, 3, 3, 2, 2,
        3, 3, 3, 4, 6, 7, 5, 4,
        3, 4, 5, 7, 6, 4, 3, 3,
        3, 2, 2, 3, 3, 3, 4, 5
        }; //new byte[32];
        
    /** Shields */
    public static int SHIELD_RADIUS = 12;
    public int shields;
    
    /** The ship singleton */
    public static Ship ship;

    private byte _explosionFrame;
    private byte _latency;                        // To limit the ship's firing power
    private byte _burst, _burstout;            // To limit the ship's firing power

    static {
        // Instantiate the ship.
        ship = new Ship();
        
        final byte[] pointx = {5,-4, -2,-4};
        final byte[] pointy = {0, 3, 0,-3};
/*
        System.out.println("3");
        // Calculate all collision detection radii (based on given ship shape)
        int i, n=0;
        double a=0, b=0, c=0, alpha=0, beta=0, gamma=0, alpha_max=0, alpha_max_old=0, x1=0, y1=0, x2=0, y2=0;
        for (i=0; i<32; i++) {
            if (i >= alpha_max) {
                System.out.println("n="+n);
                if (n<4) {x1 = pointx[n]; y1 = pointy[n++]; if (n<4) {x2 = pointx[n]; y2 = pointy[n];} else {x2 = pointx[0]; y2 = pointy[0];}}
                a = Math.sqrt((double)(x2-x1)*(x2-x1) + (y2-y1)*(y2-y1)); b = Math.sqrt((double)x2*x2 + y2*y2); c = Math.sqrt((double)x1*x1 + y1*y1);
                System.out.println("a="+a+" b="+b+" c="+c);
                alpha_max_old = alpha_max;
                alpha_max += acos((int)(256*(b*b+c*c-a*a)/(2*b*c)));
                System.out.println("alpha_max="+alpha_max);
                beta = acos((int)(256*(a*a+c*c-b*b)/(2*a*c)));
                System.out.println("beta="+beta);
            }
            alpha = i - alpha_max_old;
            gamma = 18000 - alpha - beta; while (gamma < 0) gamma += 36000; if (gamma > 36000) gamma %= 36000;
            System.out.println("gamma="+gamma);
            radii[i] = (byte)(c * Math.sin(beta) * ratioNum / Math.sin(gamma) / ratioDenom);
            System.out.println("radii="+radii[i]);
        }
        System.out.println("4");
*/
        // Calculate all ship point coordinates for all possible angles
        xcoords = new byte[128];
        ycoords = new byte[128];
        for (int i = 0; i < 32; i++) {
            int offset = i << 2;
            for (int j = 0; j < 4; j++) {
                xcoords[offset + j] = (byte)(((pointx[j] * cos[i] - pointy[j] * sin[i]) * ratioNum / ratioDenom) >> 6);
                ycoords[offset + j] = (byte)(((pointx[j] * sin[i] + pointy[j] * cos[i]) * ratioNum / ratioDenom) >> 6);
            }
        }
        radius = 5 * ratioNum / ratioDenom;
    }

    /** Empty constructor */
    public Ship() {
        shields = SHIELD_TIME[Game.field.difficulty];
    }

    /**
     * Positions the ship at its original location, orientation
     * and velocity.
     */
    public final void reset() {
        Field field = Game.field;
        angle = 0;
        moveTo(Mobile.width >> 1, Mobile.height >> 1);
        vx = 0;
        vy = 0;
        isAlive = true;
        isInvincible = true;
        countInvincible = INVINCIBLE_TIME[field.difficulty];
        teleportCounter = TELEPORT_TIME[field.difficulty];
        _burst = BURSTLENGTH[field.difficulty];
        _burstout = BURSTDELAY[field.difficulty];
        rotate(0);
    }

    /**
     * Called when the ship collides with an asteroid.
     */
    public final void explode() {
        isAlive = false;
        _explosionFrame = 0;
    }

    /**
     * Move the ship to its next position.
     */
    public final void move() {
        if (isAlive) {
            xold = x;
            yold = y;
            _x += vx;
            _y += vy;
            x = _x >> 8;
            y = _y >> 8;
            
            // If a border has been hit, wrap the trajectory around
            // the screen. The new origin is the projection of the
            // intersection point on the opposite border.
            if (x <= 0) {
                moveTo(width - 2, y);
            } else if (x >= width - 1) {
                moveTo(1, y);
            } else if (y <= 0) {
                moveTo(x, height - 2);
            } else if (y >= height - 1) {
                moveTo(x, 1);
            }

            if (_latency > 0) _latency--;
            if (_burstout > 0) _burstout--;
            if (countInvincible > 0) countInvincible--; else isInvincible = false;
            if (teleportCounter > 0) teleportCounter--;
         } else {
            _explosionFrame++;
            if (_explosionFrame > 20) {
                reset();
            }
        }
    }

    /**
     * Changes the orientation of the ship.
     */
    public final void rotate(int direction) {
        if (isAlive) {
            angle += direction;
            if (angle > 31) {
                angle -= 32;
            } else if (angle < 0) {
                angle += 32;
            }
        }
    }

    /**
     * Adds velocity to the ship.
     */
    public final void burn() {
        if (isAlive) {
            int newvx = vx + (cos[angle] << 1);
            int newvy = vy + (sin[angle] << 1);
            // clamp the ship velocity.
            if (newvx * newvx + newvy * newvy < 512*512) {
                Soundsystem sound = Game.sound;
                vx = newvx;
                vy = newvy;
                sound.playSound(sound.THRUSTERS);
            }
        }
    }

    /**
     * Decreases velocity of the ship.
     */
    public final void brake() {
        if (isAlive) {
            Soundsystem sound = Game.sound;
            int newvx = vx - (cos[angle] << 1);
            int newvy = vy - (sin[angle] << 1);
            // clamp the ship velocity.
            if (newvx * newvx + newvy * newvy < 512*512) {
                vx = newvx;
                vy = newvy;
                sound.playSound(sound.THRUSTERS);
            }
        }
    }

    /**
     * Shoots a rocket.
     */
    public final Rocket shoot(Pool rockets) {
        Field field = Game.field;
        Soundsystem sound = Game.sound;
        if (isAlive) {
            if (_latency == 0) {
                Rocket rocket = (Rocket)rockets.addNewObject();
                if (_burstout == 0) _burst = BURSTLENGTH[field.difficulty];
                _burstout = BURSTDELAY[field.difficulty];
                if (rocket != null) {
                    int offset = angle << 2;
                    sound.playSound(sound.SHOT);
                    rocket.init(xcoords[offset] + x, ycoords[offset] + y, angle, vx, vy);
                    
                    // Prevent the ship for shooting again too quickly
                    if (_burst > 0) {_latency = FIRESPEED[field.difficulty]; _burst--;} else {_burst = BURSTLENGTH[field.difficulty]; _latency = BURSTDELAY[field.difficulty]; }
                }
            }
        }
        return null;
    }

    /**
     * Ship shields on.
     */
    public final void shieldUp() {
        if (isAlive) {
            if (shields > 0) { shields--; isShieldActive = true; } else return;
        }
    }

    /**
     * Teleports the ship to another lcoation.
     */
    public final void teleport() {
        if (isAlive && teleportCounter == 0) {
            Field field = Game.field;
            Soundsystem sound = Game.sound;
            moveTo(Math.abs(Game.random.nextInt()) % width, Math.abs(Game.random.nextInt()) % height);
            isInvincible = true;
            countInvincible = INVINCIBLE_TIME[field.difficulty];
            _burst = BURSTLENGTH[field.difficulty];
            _burstout = BURSTDELAY[field.difficulty];
            sound.playSound(sound.WARP);
            teleportCounter = TELEPORT_TIME[field.difficulty];
        }
    }

    /**
     * Draws the ship at the specified location and orientation.
     */
    public static void draw(int orientation, int xpos, int ypos, Graphics g) {
        int offset = orientation << 2;
        g.drawLine(xcoords[offset] + xpos, ycoords[offset] + ypos,
                   xcoords[offset + 1] + xpos, ycoords[offset + 1] + ypos);
        g.drawLine(xcoords[offset + 1] + xpos, ycoords[offset + 1] + ypos,
                   xcoords[offset + 2] + xpos, ycoords[offset + 2] + ypos);
        g.drawLine(xcoords[offset + 2] + xpos, ycoords[offset + 2] + ypos,
                   xcoords[offset + 3] + xpos, ycoords[offset + 3] + ypos);
        g.drawLine(xcoords[offset + 3] + xpos, ycoords[offset + 3] + ypos,
                   xcoords[offset + 0] + xpos, ycoords[offset + 0] + ypos);
        if (Ship.ship.isShieldActive) {
            g.setColor(Field.COLOR_SHIELD);
            int radius = SHIELD_RADIUS; 
            g.drawArc(xpos - radius, ypos - radius, radius << 1, radius << 1, 0, 360);
        }
    }

    /**
     * Draws the ship in the specified graphic context.
     */
    public final void draw(Graphics g) {
        if (isAlive) {
            draw(angle, x, y, g);
        } else {
            if (_explosionFrame < 6) {
                g.setColor(Field.COLOR_EXPLOSION);
                int radius = _explosionFrame;
                g.drawArc(x - radius, y - radius, radius << 1, radius << 1, 0, 360);
                radius = _explosionFrame + 2;
                g.drawArc(x - radius, y - radius, radius << 1, radius << 1, 0, 360);
                radius = _explosionFrame + 4;
                g.drawArc(x - radius, y - radius, radius << 1, radius << 1, 0, 360);
            }
        }
    }
}

