/**
 * Copyright 2005 Simon Lehmayr
 *
 * This file is part of Asteroid Zone. Asteroid Zone is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 * Asteroid Zone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Asteroid Zone; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 */

package asteroids;

import javax.microedition.midlet.*;
import java.util.*;
// Required for sound
import javax.microedition.media.*;
import java.io.*;

/**
 * The game sound system running in an extra thread.
 * @author Jean-Francois Doue
 * @author Pavel Machek
 * @author Simon Lehmayr
 * @version 1.6, 2005/12/20
 */
public class Soundsystem extends Thread {

    /** Game sounds */
    public Player crashSound;
    public Player explosionSound;
    public Player fireSound;
    public Player missileSound;
    public Player ufoSound;
    public Player thrustersSound;
    public Player warpSound;
    public Player musicSound;

    /** Constants */
    public static final int INIT = 0;
    public static final int READY = 1;
    public static final int JOBPLAY = 3;
    public static final int JOBSTOP = 4;
    public static final int JOBMUTE = 5;
    public static final int JOBMUSICMUTE = 6;

    public static final int PRIOCRASH = 10;
    public static final int PRIOMISSILE = 9;
    public static final int PRIOUFO = 8;
    public static final int PRIOEXPLOSION = 7;
    public static final int PRIOWARP = 6;
    public static final int PRIOTHRUSTERS = 5;
    public static final int PRIOSHOT = 1;

    public static final int CRASH = 10;
    public static final int MISSILE = 9;
    public static final int UFO = 8;
    public static final int EXPLOSION = 7;
    public static final int WARP = 6;
    public static final int THRUSTERS = 5;
    public static final int SHOT = 1;

    /** States */
    public static boolean running = true;
    public boolean mute=true, musicMute=true, ready=false;
    public int state=0;
    public Player selectedPlayer, lastPlayer;

    /** Constructor */
    public Soundsystem () {
    }

    /** Sound loader */
    private void loadSounds() {
        try {
            InputStream is_crash = getClass().getResourceAsStream("/explosion.wav");
            InputStream is_explosion = getClass().getResourceAsStream("/crash.wav");
            InputStream is_fire = getClass().getResourceAsStream("/fire.wav");
            InputStream is_missile = getClass().getResourceAsStream("/missile.wav");
            InputStream is_ufo = getClass().getResourceAsStream("/saucer.wav");
            InputStream is_thrusters = getClass().getResourceAsStream("/thrusters.wav");
            InputStream is_warp = getClass().getResourceAsStream("/warp.wav");
            InputStream is_music = getClass().getResourceAsStream("/music.mid");
            crashSound = Manager.createPlayer(is_crash, "audio/x-wav");
            explosionSound = Manager.createPlayer(is_explosion, "audio/x-wav");
            fireSound = Manager.createPlayer(is_fire, "audio/x-wav");
            missileSound = Manager.createPlayer(is_missile, "audio/x-wav");
            ufoSound = Manager.createPlayer(is_ufo, "audio/x-wav");
            thrustersSound = Manager.createPlayer(is_thrusters, "audio/x-wav");
            warpSound = Manager.createPlayer(is_warp, "audio/x-wav");
            musicSound = Manager.createPlayer(is_music, "audio/midi");
            musicSound.setLoopCount(1);

            musicSound.realize(); musicSound.prefetch();
            crashSound.realize(); crashSound.prefetch();
            explosionSound.realize(); explosionSound.prefetch();
            fireSound.realize(); fireSound.prefetch();
            missileSound.realize(); missileSound.prefetch();
            ufoSound.realize(); ufoSound.prefetch();
            thrustersSound.realize(); thrustersSound.prefetch();
            warpSound.realize(); warpSound.prefetch();
        }
        catch (MediaException pe) {}
        catch (IOException ioe) {}
    }

    /**
     * Plays the given sound.
     * Stops the player and restarts it to get always a fitting sound.
     */
    public void playSound_t(Player p) {
        if (mute || !ready) return;
        try {
            if (lastPlayer.getState() == Player.STARTED) 
                lastPlayer.stop();
            lastPlayer = p;
            if (p.getState() == Player.STARTED) 
                p.stop();
            p.setMediaTime(0);
            p.start();
        }
        catch (MediaException pe) {}
    }

    public void playSound(int p) {
        if (mute || !ready) return;
//        if (state != READY) return;
        switch (p) {
            case CRASH: selectedPlayer = crashSound; break;
            case EXPLOSION: selectedPlayer = explosionSound; break;
            case MISSILE: selectedPlayer = missileSound; break;
            case UFO: selectedPlayer = ufoSound; break;
            case THRUSTERS: selectedPlayer = thrustersSound; break;
            case WARP: selectedPlayer = warpSound; break;
            case SHOT: selectedPlayer = fireSound; break;
        }
        state = JOBPLAY;
    }

    /**
     * Sound switch.
     * Stops all sounds when turned off.
     */
    public void setMute_t() {
        if (!ready) return;
        try {
        if (mute) {
            musicSound.stop();
            crashSound.stop();
            explosionSound.stop();
            fireSound.stop();
            ufoSound.stop();
            missileSound.stop();
            thrustersSound.stop();
            warpSound.stop();
        }
        } catch (MediaException pe) {}
    }
    
    public void setMute(boolean m) {
        if (!ready) return;
        mute = m;
        state = JOBMUTE;
    }

    /**
     * Music switch.
     * Stops the music when turned off.
     */
    public void setMusicMute_t() {
        if (!ready) return;
        try {
        if (musicMute)
            musicSound.stop();
        else {
            musicSound.setLoopCount(-1);
            musicSound.start();
        }
        } catch (MediaException pe) {}
    }

    public void setMusicMute(boolean m) {
        if (!ready) return;
        musicMute = m;
        state = JOBMUSICMUTE;
    }

    public void run() {
        while (running) {
        try {
            long t1 = System.currentTimeMillis();
            if (state == INIT) {
                loadSounds();
                lastPlayer = crashSound;
                ready = true;
                state = READY;
            }
            else if (state == JOBPLAY) {
                playSound_t(selectedPlayer);
                state = READY;
            }
            else if (state == JOBMUTE) {
                setMute_t();
                state = READY;
            }
            else if (state == JOBMUSICMUTE) {
                setMusicMute_t();
                state = READY;
            }
            long t2 = System.currentTimeMillis();
            Thread.currentThread().sleep(Math.max(Game.field.GAME_SPEED[Game.field.difficulty] - (t2-t1), 20));
        }
        catch(java.lang.InterruptedException e) {}
    }
    }
}
